@@ -28,6 +28,7 @@ my $SERVICE_START = 0;
 my $SERVICE_STOP  = 1;
 
 use Init::ServicesList;
+use Utils::Report;
 
 sub get_runlevels
 {
@@ -49,16 +50,17 @@ sub get_runlevels
      "suse-9.0"         => "redhat-6.2",
      "pld-1.0"          => "redhat-6.2",
      "vine-3.0"         => "redhat-6.2",
-     "slackware-9.1.0"  => "slackware-9.1.0",
+     "slackware-9.1.0"  => "freebsd-5",
      "gentoo"           => "gentoo",
-     "freebsd-5"        => "slackware-9.1.0",
+     "archlinux"        => "freebsd-5",
+     "freebsd-5"        => "freebsd-5",
      "solaris-2.11"     => "freebsd-5",
     );
 
   %runlevels=
     (
      "redhat-6.2"      => [ "0", "1", "2", "3", "4", "5", "6" ],
-     "gentoo"          => [ "boot", "default", "nonetwork" ],
+     "gentoo"          => &get_gentoo_runlevels (),
      "freebsd-5"       => [ "default" ],
     );
 
@@ -506,7 +508,8 @@ sub set_filerc_services
 sub get_bsd_scripts_list
 {
   my ($files) = [ "rc.M", "rc.inet2", "rc.4" ];
-  my ($file, $i, @scripts);
+  my ($file, $i, %scripts);
+  my ($service, $name);
 
   foreach $i (@$files)
   {
@@ -515,7 +518,7 @@ sub get_bsd_scripts_list
 
     if (!$fd) {
       &Utils::Report::do_report ("rc_file_read_failed", $file);
-      return undef;
+      next;
     }
 
     while (<$fd>)
@@ -525,54 +528,50 @@ sub get_bsd_scripts_list
       if ($line =~ /^if[ \t]+\[[ \t]+\-x[ \t]([0-9a-zA-Z\/\.\-_]+) .*\]/)
       {
         $service = $1;
+        $name = $service;
+        $name =~ s/^.*\///;
+        $name =~ s/^rc\.//;
 
-        push @scripts, $service;
+        $scripts{$name} = $service;
       }
     }
 
     &Utils::File::close_file ($fd);
   }
 
-  return \@scripts;
+  return \%scripts;
+}
+
+sub get_bsd_service_status
+{
+  my ($service) = @_;
+  return (-x $service) ? $SERVICE_START : $SERVICE_STOP;
 }
 
 sub get_bsd_service_info
 {
-  my ($service) = @_;
-  my ($script, @runlevels);
+  my ($service, $name) = @_;
+  my (@runlevels, $status);
 
-  $script = $service;
-  $script =~ s/^.*\///;
-  $script =~ s/^rc\.//;
+  return if (! Utils::File::exists ($service));
 
-  return undef if (! Utils::File::exists ($service));
+  $status = &get_bsd_service_status ($service);
+  push @runlevels, [ "default", $status, 0 ];
 
-  # hardcode the fourth runlevel, it's the graphical one
-  if ( -x $service)
-  {
-    push @runlevels, [ "4", $SERVICE_START, 0 ];
-  }
-  else
-  {
-    push @runlevels, [ "4", $SERVICE_STOP, 0 ];
-  }
-
-  # FIXME: $service contains full path, would
-  # be nice to just show the $script name
-  return ($service, $runlevels);
+  return ($name, \@runlevels);
 }
 
 sub get_bsd_services
 {
-  my (@arr, @scripts, $script);
+  my (@arr, %scripts, $name);
 
   $scripts = &get_bsd_scripts_list ();
 
-  foreach $script (@scripts)
+  foreach $name (keys %$scripts)
   {
     my (@info);
 
-    @info = &get_bsd_service_info ($script);
+    @info = &get_bsd_service_info ($$scripts{$name}, $name);
     push @arr, \@info if (scalar (@info));
   }
 
@@ -587,13 +586,13 @@ sub run_bsd_script
   return if (!&Utils::File::exists ($service));
 
   # if it's not executable then chmod it
-  if (!((stat ($service))[2] & (S_IXUSR || S_IXGRP || S_IXOTH)))
+  if (! -x $service)
   {
     $chmod = 1;
     &Utils::File::run ("chmod ugo+x $service");
   }
-  
-  &Utils::File::run ("$service $arg");
+
+  &Utils::File::run ("$service $arg", 1);
 
   # return it to it's normal state
   if ($chmod)
@@ -603,41 +602,28 @@ sub run_bsd_script
 }
 
 # This function stores the configuration in a bsd init
-sub get_bsd_service_path
-{
-  my ($script) = @_;
-  my (@scripts, $elem, $i);
-
-  $scripts = &get_bsd_scripts_list ();
-
-  foreach $i (@scripts)
-  {
-    $elem = $i;
-    $elem =~ s/^.*\///;
-    $elem =~ s/^rc\.//;
-
-    return $i if ($elem eq $script);
-  }
-
-  return undef;
-}
-
 sub set_bsd_services
 {
   my ($services) = @_;
-  my ($script, $runlevels, @scripts);
+  my ($script, $runlevels, %scripts);
+  my ($status);
+
+  $scripts = &get_bsd_scripts_list ();
 
   foreach $service (@$services)
   {
-    $script = &get_bsd_service_path ($$service[0]);
+    $script = $$scripts{$$service[0]};
     $runlevels = $$service[1];
     $runlevel  = $$runlevels[0];
 
     next if ($script eq undef);
 
-    $action = $$runlevel[1];
+    $status = $$runlevel[1];
+    $status = $SERVICE_STOP if ($status eq undef);
 
-    if ($action == $SERVICE_START)
+    next if ($status == &get_bsd_service_status ($script));
+
+    if ($status == $SERVICE_START)
     {
       &Utils::File::run ("chmod ugo+x $script");
       &run_bsd_script ($script, "start");
@@ -654,42 +640,55 @@ sub set_bsd_services
 sub get_gentoo_service_status
 {
   my ($script, $runlevel) = @_;
-  my ($services) = &get_gentoo_services_by_runlevel ($runlevel);
+  my ($services) = &get_gentoo_services_for_runlevel ($runlevel);
 
-  foreach $i (@$services)
-  {
-    return 1 if ($i eq $script);
-  }
-
-  return 0;
+  return ($$services {$script});
 }
 
 sub get_gentoo_runlevels
 {
-  my($raw_output) = Utils::File::run_backtick("rc-status -l");
-  my(@runlevels) = split(/\n/,$raw_output);
+  my($raw_output) = Utils::File::run_backtick("rc-status -nc -l");
+  my(@runlevels);
 
-  return @runlevels;
+  return undef if (!$raw_output);
+  @runlevels = split(/\n/,$raw_output);
+
+  return \@runlevels;
 }
 
-sub get_gentoo_services_by_runlevel
+sub get_gentoo_services_for_runlevel
 {
   my($runlevel) = @_;
-  my($raw_output) = Utils::File::run_backtick("rc-status $runlevel");
+  my($raw_output) = Utils::File::run_backtick("rc-status -nc $runlevel");
   my(@raw_lines) = split(/\n/,$raw_output);
-  my(@services);
-  my($line);
+  my($line, $service);
+  my(%services);
 
   foreach $line (@raw_lines)
   {
     if ($line !~ /^Runlevel/)
     {
-      $line=(split(" ",$line))[0];
-	    push(@services,$line);
+      $service = (split(" ",$line))[0];
+      $services{$service} = 1;
 	  }
   }
 
-  return \@services
+  return \%services
+}
+
+sub get_gentoo_runlevels_services
+{
+  my (%runlevels_services, $runlevels);
+
+  $runlevels = &get_gentoo_runlevels ();
+  return undef if (!$runlevels);
+
+  foreach $runlevel (@$runlevels)
+  {
+    $runlevels_services{$runlevel} = &get_gentoo_services_for_runlevel ($runlevel);
+  }
+
+  return \%runlevels_services;
 }
 
 sub get_gentoo_services_list
@@ -721,50 +720,16 @@ sub gentoo_service_exists
   return 0;
 }
 
-sub get_gentoo_runlevels_by_service
+sub get_gentoo_runlevels_status
 {
-  my ($service) = @_;
-  my(@runlevels,@services_in_runlevel,@contain_runlevels, $runlevel);
-  my ($elem);
+  my ($service, $runlevels_services) = @_;
+  my (@arr, $services_in_runlevel);
 
-  # let's do some caching to improve performance
-  if ($gentoo_services_hash eq undef)
+  foreach $runlevel (keys %$runlevels_services)
   {
-    @runlevels = &get_gentoo_runlevels ();
+    $services_in_runlevel = $$runlevels_services {$runlevel};
 
-    foreach $runlevel (@runlevels)
-    {
-      $$gentoo_services_hash{$runlevel} = &get_gentoo_services_by_runlevel ($runlevel);
-    }
-  }
-
-  if (&gentoo_service_exists($service))
-  {
-    foreach $runlevel (keys %$gentoo_services_hash)
-    {
-      $services_in_runlevel = $$gentoo_services_hash {$runlevel};
-
-      foreach $elem (@$services_in_runlevel)
-      {
-        push (@contain_runlevels, $runlevel) if ($elem eq $service);
-      }
-    }
-  }
-
-  return @contain_runlevels;
-}
-
-sub get_gentoo_runlevel_status_by_service
-{
-  my ($service) = @_;
-  my (@arr, @ret);
-  my (@runlevels) = &get_gentoo_runlevels();
-  my (@started) = &get_gentoo_runlevels_by_service($service);
-  my (%start_runlevels) = map { $started[$_], 1 } 0 .. $#started;
-
-  foreach $runlevel (@runlevels)
-  {
-    if (defined $start_runlevels{$runlevel})
+    if ($$services_in_runlevel{$service})
     {
       push @arr, [ $runlevel, $SERVICE_START, 0 ];
     }
@@ -779,26 +744,25 @@ sub get_gentoo_runlevel_status_by_service
 
 sub get_gentoo_service_info
 {
-	my ($service) = @_;
-	my ($script, @actions, @runlevels);
-	my ($role);
-	
-  $runlevels = &get_gentoo_runlevel_status_by_service ($service);
+  my ($service, $runlevels_services) = @_;
+  my (@runlevels_info);
 
-  return ($service, $runlevels);
+  $runlevels_info = &get_gentoo_runlevels_status ($service, $runlevels_services);
+
+  return ($service, $runlevels_info);
 }
 
 sub get_gentoo_services
 {
-  my ($service);
-  my (@arr);
+  my ($service, @arr);
   my ($service_list) = &get_gentoo_services_list ();
+  my ($runlevels_services) = &get_gentoo_runlevels_services ();
 
   foreach $service (@$service_list)
   {
     my (@info);
 
-    @info = &get_gentoo_service_info ($service);
+    @info = &get_gentoo_service_info ($service, $runlevels_services);
     push @arr, \@info if scalar (@info);
   }
 
@@ -829,7 +793,14 @@ sub run_gentoo_script
 
 sub set_gentoo_service_status
 {
-  my ($script, $rl, $action) = @_;
+  my ($script, $rl, $status, $runlevels_services) = @_;
+  my ($services_in_runlevel, $old_status);
+
+  $services_in_runlevel = $$runlevels_services {$rl};
+  $old_status = ($$services_in_runlevel{$script}) ?
+      $SERVICE_START : $SERVICE_STOP;
+
+  return if ($status == $old_status);
 
   if ($action == $SERVICE_START)
   {
@@ -848,6 +819,9 @@ sub set_gentoo_services
 {
   my ($services) = @_;
   my ($action, $rl, $script, $arr);
+  my ($runlevels_services) = &get_gentoo_runlevels_services ();
+
+  return if (!$runlevels_services);
 
   foreach $service (@$services)
   {
@@ -858,7 +832,8 @@ sub set_gentoo_services
     {
       $action = $$i[1];
       $rl = $$i[0];
-      &set_gentoo_service_status ($script, $rl, $action);
+      &set_gentoo_service_status ($script, $rl, $action,
+                                  $runlevels_services);
     }
   }
 }
@@ -872,7 +847,7 @@ sub get_rcng_status_by_service
   # This is the only difference between rcNG and archlinux
   if ($Utils::Backend::tool{"platform"} eq "archlinux")
   {
-      return &Utils::File::exists ("/var/run/daemons/$service");
+    return &Utils::File::exists ("/var/run/daemons/$service");
   }
   else
   {
@@ -908,7 +883,7 @@ sub get_rcng_service_info
     push @runlevels, [ "default", $SERVICE_STOP, 0 ];
   }
 
-  return ($script, $runlevels);
+  return ($script, \@runlevels);
 }
 
 sub get_rcng_services
@@ -936,13 +911,13 @@ sub run_rcng_script
 
   if (!&Utils::File::run ("/etc/rc.d/$service $arg"))
   {
-    &Utils::File::do_report ("service_sysv_op_success", $service, $arg);
-    &Utils::File::leave ();
+    &Utils::Report::do_report ("service_sysv_op_success", $service, $arg);
+    &Utils::Report::leave ();
     return 0;
   }
 
-  &Utils::File::do_report ("service_sysv_op_failed", $service, $arg);
-  &Utils::File::leave ();
+  &Utils::Report::do_report ("service_sysv_op_failed", $service, $arg);
+  &Utils::Report::leave ();
   return -1;
 }
 
@@ -1005,17 +980,29 @@ sub set_archlinux_service_status
   my ($daemons);
 
   $daemons = &Utils::Parse::get_sh ($rcconf, "DAEMONS");
+  $daemons =~ s/[\(\)]//g;
 
-  if (($daemons =~ m/$script/) && !$active)
+  # escape these chars
+  $script =~ s/([\\\.\^\$\*\+\?\{\}\[\]\(\)\|])/\\\1/g;
+  $notscript = "\!" . $script;
+
+  if (($daemons =~ m/$notscript/) && $active)
   {
-    $daemons =~ s/$script[ \t]*//;
+    # It was disabled, enable it
+    $daemons =~ s/$notscript/$script/g;
+  }
+  elsif (($daemons =~ m/$script/) && !$active)
+  {
+    # It was enabled, disable it
+    $daemons =~ s/$script/$notscript/g;
   }
   elsif (($daemons !~ m/$script/) && $active)
   {
-    $daemons =~ s/network/network $script/g;
+    $daemons .= " ".$script;
   }
 
-  &Utils::Replace::set_sh ($rcconf, "DAEMONS", $daemons);
+  $daemons = "(" . $daemons . ")";
+  &Utils::Replace::set_sh ($rcconf, "DAEMONS", $daemons, 1);
   &run_rcng_script ($service, ($active) ? "start" : "stop");
 }
 
@@ -1050,7 +1037,7 @@ sub get_suse_service_info ($service)
 {
   my ($service) = @_;
   my (@runlevels, $link, $runlevel);
-                                                                                                                                                             
+
   foreach $link (<$rcd_path/rc[0-9S].d/S[0-9][0-9]$service>)
   {
     $link =~ s/$rcd_path\///;
@@ -1304,7 +1291,7 @@ sub get_init_type
   {
     return "rcng";
   }
-  elsif (($gst_dist =~ /gentoo/) || ($gst_dist =~ /^vlos/))
+  elsif ($gst_dist =~ /gentoo/)
   {
     return "gentoo";
   }
